"""Async database interaction code relating to Genesets."""

from typing import List, Optional

from geneweaver.core.enum import GeneIdentifier, GenesetTier, Species
from geneweaver.db.query import geneset as geneset_query
from psycopg import AsyncCursor
from psycopg.rows import Row


async def get(
    cursor: AsyncCursor,
    gs_id: Optional[int] = None,
    owner_id: Optional[int] = None,
    curation_tier: Optional[GenesetTier] = None,
    species: Optional[Species] = None,
    name: Optional[str] = None,
    abbreviation: Optional[str] = None,
    publication_id: Optional[int] = None,
    pubmed_id: Optional[int] = None,
    gene_id_type: Optional[GeneIdentifier] = None,
    search_text: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get genesets from the database.

    :param cursor: An async database cursor.
    :param gs_id: Show only results with this geneset ID.
    :param owner_id: Show only results owned by this user ID.
    :param curation_tier: Show only results of this curation tier.
    :param species: Show only results associated with this species.
    :param name: Show only results with this name.
    :param abbreviation: Show only results with this abbreviation.
    :param publication_id: Show only results with this publication ID (internal).
    :param pubmed_id: Show only results with this PubMed ID.
    :param gene_id_type: Show only results with this gene ID type.
    :param search_text: Return genesets that match this search text (using PostgreSQL
                        full-text search).
    :param limit: Limit the number of results.
    :param offset: Offset the results.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *geneset_query.get(
            is_readable_by=is_readable_by,
            gs_id=gs_id,
            owner_id=owner_id,
            curation_tier=curation_tier,
            species=species,
            name=name,
            abbreviation=abbreviation,
            publication_id=publication_id,
            pubmed_id=pubmed_id,
            gene_id_type=gene_id_type,
            search_text=search_text,
            limit=limit,
            offset=offset,
            with_publication_info=with_publication_info,
        )
    )

    return await cursor.fetchall()


async def by_id(
    cursor: AsyncCursor,
    gs_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> Optional[Row]:
    """Get a geneset by its ID.

    :param cursor: An async database cursor.
    :param gs_id: The geneset ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.
    """
    await cursor.execute(
        *geneset_query.get(
            gs_id=gs_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )

    return await cursor.fetchone()


async def by_owner_id(
    cursor: AsyncCursor,
    owner_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get genesets by owner ID.

    :param cursor: An async database cursor.
    :param owner_id: The owner ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *geneset_query.get(
            owner_id=owner_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )
    return await cursor.fetchall()


async def by_project_id(
    cursor: AsyncCursor,
    project_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get a list of genesets by their membership in a project.

    :param cursor: An async database cursor.
    :param project_id: The project ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *geneset_query.by_project_id(
            project_id=project_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )
    return await cursor.fetchall()



"""Utility functions for the geneset query."""

from typing import Optional, Set, Tuple, Union

from geneweaver.core.enum import GenesetTier
from geneweaver.db.query.geneset.const import (
    GENESET_FIELDS,
    GENESET_TSVECTOR,
    PUB_FIELDS,
)
from geneweaver.db.query.search import utils
from geneweaver.db.query.utils import (
    ParamDict,
    SQLList,
)
from psycopg.sql import SQL, Composed

GenesetTierOrTiers = Union[GenesetTier, Set[GenesetTier]]


def format_select_query(
    with_publication_info: bool = False,
    with_publication_join: bool = False,
) -> Composed:
    """Format the geneset query.

    Construct the geneset query with appropriate fields and joins.

    If the `with_publication_info` flag is set to True, the query will include the
    publication table for querying AND return data. When `with_publication_info` is set
    to True, the `with_publication_join` option is ignored.

    If the `with_publication_join` flag is set to True, the query will include
    the publication table for querying, but not return. This option is ignored if the
    `with_publication_info` flag is set to True.

    :param with_publication_info: Whether to include publication info in return.
    :param with_publication_join: Whether to include publication for querying.

    :return: The formatted query.
    """
    query = SQL("SELECT")
    if with_publication_info:
        query = (
            query
            + SQL(",").join(GENESET_FIELDS + PUB_FIELDS)
            + SQL("FROM geneset LEFT OUTER JOIN publication")
            + SQL("ON geneset.pub_id = publication.pub_id")
        )
    elif with_publication_join:
        query = (
            query
            + SQL(",").join(GENESET_FIELDS)
            + SQL("FROM geneset LEFT OUTER JOIN publication")
            + SQL("ON geneset.pub_id = publication.pub_id")
        )
    else:
        query = query + SQL(",").join(GENESET_FIELDS) + SQL("FROM geneset")
    return query


def is_readable(
    existing_filters: SQLList,
    existing_params: ParamDict,
    is_readable_by: Optional[int] = None,
) -> Tuple[SQLList, ParamDict]:
    """Add the is_readable filter to the query.

    :param existing_filters: The existing filters.
    :param existing_params: The existing parameters.
    :param is_readable_by: The user ID to filter by.
    """
    if is_readable_by is not None:
        existing_filters.append(
            SQL("production.geneset_is_readable2(%(is_readable_by)s, geneset.gs_id)")
        )
        existing_params["is_readable_by"] = is_readable_by
    return existing_filters, existing_params


def search(
    existing_filters: SQLList,
    existing_params: ParamDict,
    search_text: Optional[str] = None,
) -> Tuple[SQLList, ParamDict]:
    """Add the search filter to the query.

    :param existing_filters: The existing filters.
    :param existing_params: The existing parameters.
    :param search_text: The search text to filter by.
    """
    if search_text is not None:
        search_sql, search_params = utils.search_query(GENESET_TSVECTOR, search_text)
        existing_filters.append(search_sql)
        existing_params.update(search_params)
    return existing_filters, existing_params


def restrict_tier(
    existing_filters: SQLList,
    existing_params: ParamDict,
    curation_tier: Optional[GenesetTierOrTiers] = None,
) -> Tuple[SQLList, ParamDict]:
    """Restrict the query by curation tier.

    :param existing_filters: The existing filters.
    :param existing_params: The existing parameters.
    :param curation_tier: The curation tier to filter by.
    """
    if isinstance(curation_tier, GenesetTier):
        curation_tier = {curation_tier}
    if curation_tier is not None:
        existing_filters.append(SQL("geneset.cur_id = ANY(%(curation_tier)s)"))
        existing_params["curation_tier"] = [int(tier) for tier in curation_tier]
    return existing_filters, existing_params



"""Functions for generating search queries."""

from enum import Enum
from typing import Tuple

from psycopg.sql import SQL, Composed


class SearchConfig(Enum):
    """Search configuration options."""

    SIMPLE = "simple"
    ENGLISH = "english"

    def __str__(self: "SearchConfig") -> str:
        """Render the value as a string."""
        return self.value


class QueryType(Enum):
    """Postgresql Query Type."""

    PLAINTO = "plainto_tsquery"
    PHRASETO = "phraseto_tsquery"
    WEBSEARCH = "websearch_to_tsquery"

    def __str__(self: "SearchConfig") -> str:
        """Render the value as a string."""
        return self.value


SEARCH_QUERIES = {
    QueryType.PLAINTO: SQL("@@ plainto_tsquery({search_config}, %(search)s)"),
    QueryType.PHRASETO: SQL("@@ phraseto_tsquery({search_config}, %(search)s)"),
    QueryType.WEBSEARCH: SQL("@@ websearch_to_tsquery({search_config}, %(search)s)"),
}


def search_query(
    search_column: Composed,
    search_string: str,
    search_config: SearchConfig = SearchConfig.SIMPLE,
    query_type: QueryType = QueryType.WEBSEARCH,
) -> Tuple[Composed, dict]:
    """Generate a search query.

    Works generically with any table that has a search column.

    :param search_column: The column to search.
    :param search_string: The string to search for.
    :param search_config: The search configuration to use.
    :param query_type: The query type to use.
    """
    query = SEARCH_QUERIES[query_type].format(search_config=str(search_config))
    return search_column + query, {"search": search_string}



"""Query generation functions for projects."""

from typing import Optional, Tuple

from psycopg.sql import Composed


def get(
    project_id: Optional[int] = None,
    owner_id: Optional[int] = None,
    name: Optional[str] = None,
    starred: Optional[bool] = None,
    search_text: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Tuple[Composed, dict]:
    """Get projects by any filtering criteria.

    NOTE: NOT IMPLEMENTED
    """
    raise NotImplementedError()


def shared_with_user(
    user_id: int,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Tuple[Composed, dict]:
    """Get project that are shared with a user.

    NOTE: NOT IMPLEMENTED
    """
    raise NotImplementedError()


def add(
    user_id: int,
    name: str,
    notes: str,
    starred: bool = False,
) -> Tuple[Composed, dict]:
    """Add a new publication.

    NOTE: NOT IMPLEMENTED
    """
    raise NotImplementedError()



"""Geneset thresholding database functions."""

from geneweaver.core.schema.score import GenesetScoreType
from geneweaver.db.query import threshold as threshold_query
from psycopg import Cursor


def set_geneset_threshold(
    cursor: Cursor,
    geneset_id: int,
    geneset_score_type: GenesetScoreType,
) -> None:
    """Update the threshold of a geneset and its values.

    :param cursor: An database cursor.
    :param geneset_id: The ID of the geneset to update.
    :param geneset_score_type: The score threshold to set.
    :return: Nothing.
    """
    cursor.execute(
        *threshold_query.set_geneset_threshold(geneset_id, geneset_score_type)
    )



"""Species database functions."""

from typing import List, Optional

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.db.query import species as species_query
from psycopg import Cursor, rows


def get(
    cursor: Cursor,
    taxonomic_id: Optional[int] = None,
    reference_gene_db_id: Optional[GeneIdentifier] = None,
    species: Optional[Species] = None,
) -> List:
    """Get species info.

    :param cursor: The database cursor.
    :param taxonomic_id: The taxonomic id.
    :param reference_gene_db_id: The reference gene database id.
    :param species: The species id.

    :return: All species that match the queries.
    """
    cursor.execute(
        *species_query.get(
            taxonomic_id=taxonomic_id,
            reference_gene_db_id=reference_gene_db_id,
            species=species,
        )
    )

    return cursor.fetchall()


def get_by_id(
    cursor: Cursor,
    species: Species,
) -> Optional[rows.Row]:
    """Get species info by species id.

    :param cursor: The database cursor.
    :param species: The species enum to query info for.
    :return: The species info for the provided enum.
    """
    cursor.execute(*species_query.get(species=species))

    return cursor.fetchone()



"""Generate SQL queries for publications."""

from typing import Iterable, Optional, Tuple

from geneweaver.db.utils import format_sql_fields
from psycopg import rows
from psycopg.sql import SQL, Composed, Identifier, Placeholder

PUB_FIELD_MAP = {
    "pub_id": "id",
    "pub_authors": "authors",
    "pub_title": "title",
    "pub_abstract": "abstract",
    "pub_journal": "journal",
    "pub_volume": "volume",
    "pub_pages": "pages",
    "pub_month": "month",
    "pub_year": "year",
    "pub_pubmed": "pubmed_id",
}

PUB_FIELDS = format_sql_fields(PUB_FIELD_MAP, query_table="publication")

PUB_QUERY = SQL("SELECT") + SQL(",").join(PUB_FIELDS) + SQL("FROM publication")

PUB_INSERT_COLS = SQL(",").join(
    [Identifier(k) for k in PUB_FIELD_MAP.keys() if k != "pub_id"]
)
PUB_INSERT_VALS = SQL(",").join(
    [Placeholder(k) for k in PUB_FIELD_MAP.keys() if k != "pub_id"]
)


def get(
    pub_id: Optional[int] = None,
    authors: Optional[str] = None,
    title: Optional[str] = None,
    abstract: Optional[str] = None,
    journal: Optional[str] = None,
    volume: Optional[str] = None,
    pages: Optional[str] = None,
    month: Optional[str] = None,
    year: Optional[str] = None,
    pubmed: Optional[str] = None,
    search_text: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Tuple[Composed, dict]:
    """Get publications by some criteria.

    NOTE: NOT IMPLEMENTED
    """
    raise NotImplementedError()


def by_id(pub_id: int) -> Optional[rows.Row]:
    """Create a psycopg query to get a publication by ID.

    :param pub_id: The publication ID (geneweaver internal) to search for.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (PUB_QUERY + SQL("WHERE pub_id = %(pub_id)s")).join(" ")
    params = {"pub_id": pub_id}
    return query, params


def by_geneset_id(geneset_id: int) -> Tuple[Composed, dict]:
    """Create a psycopg query to get a publication by geneset ID.

    :param geneset_id: The geneset ID to search for.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (
        PUB_QUERY
        + SQL("JOIN geneset ON publication.pub_id = geneset.pub_id")
        + SQL("WHERE gs_id = %(geneset_id)s")
    ).join(" ")

    params = {"geneset_id": geneset_id}
    return query, params


def by_pubmed_id(pubmed_id: int) -> Tuple[Composed, dict]:
    """Create a psycopg query to get a publication by PubMed ID.

    :param pubmed_id: The PubMed ID to search for.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (PUB_QUERY + SQL("WHERE pub_pubmed = %(pmid)s")).join(" ")
    # PubMed IDs are integers, but are stored as strings in the database.
    params = {"pmid": str(pubmed_id)}
    return query, params


def by_pubmed_ids(pubmed_ids: Iterable[int]) -> Tuple[Composed, dict]:
    """Create a psycopg query to get publications by a list of PubMed IDs.

    :param pubmed_ids: The PubMed IDs to search for.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (PUB_QUERY + SQL("WHERE pub_pubmed = ANY(%(pubmed_ids)s)")).join(" ")
    params = {"pubmed_ids": list(pubmed_ids)}
    return query, params


def add(
    authors: str,
    title: str,
    abstract: str,
    journal: str,
    pubmed_id: str,
    volume: Optional[str] = None,
    pages: Optional[str] = None,
    month: Optional[str] = None,
    year: Optional[int] = None,
) -> Tuple[Composed, dict]:
    """Create a psycopg query to add a publication to the database.

    :param authors: The authors of the publication.
    :param title: The title of the publication.
    :param abstract: The abstract of the publication.
    :param journal: The journal of the publication.
    :param volume: The volume of the publication.
    :param pages: The pages of the publication.
    :param month: The month of the publication.
    :param year: The year of the publication.
    :param pubmed_id: The PubMed ID of the publication.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (
        SQL("INSERT INTO publication")
        + SQL("(")
        + PUB_INSERT_COLS
        + SQL(")")
        + SQL("VALUES")
        + SQL("(")
        + PUB_INSERT_VALS
        + SQL(")")
        + SQL("RETURNING pub_id")
    ).join(" ")

    params = {
        "pub_authors": authors,
        "pub_title": title,
        "pub_abstract": abstract,
        "pub_journal": journal,
        "pub_volume": volume,
        "pub_pages": pages,
        "pub_month": month,
        "pub_year": year,
        "pub_pubmed": pubmed_id,
    }

    return query, params



"""The settings class definition for the GeneWeaver Database module.

This class is intended to be used with environment variables, and will attempt to
load them from a `.env` file in the root of the project. The following environment
variables are used:
GWDB_SERVER=your_server
GWDB_USERNAME=your_user
GWDB_PASSWORD=your_password
GWDB_NAME=your_database_name
"""

# ruff: noqa: N805, ANN101, ANN401
from typing import Any, Dict, Optional

from pydantic import BaseSettings, PostgresDsn, validator


class Settings(BaseSettings):
    """Settings class for the GeneWeaver Database module."""

    DEBUG_MODE = False

    CONNECTION_SCHEME: str = "postgresql"

    SERVER: str
    USERNAME: str
    PASSWORD: str = ""
    NAME: str = ""
    PORT: int = 5432
    URI: Optional[str] = None

    @validator("SERVER", pre=True)
    def replace_localhost(cls, v: str) -> str:
        """Replace localhost with 127.0.0.1."""
        if v == "localhost":
            return "127.0.0.1"
        return v

    @validator("URI", pre=True)
    def assemble_db_connection(cls, v: Optional[str], values: Dict[str, Any]) -> str:
        """Build the database connection string, unless one is provided."""
        if isinstance(v, str):
            return v
        return str(
            PostgresDsn.build(
                scheme=values.get("CONNECTION_SCHEME"),
                user=values.get("USERNAME"),
                password=values.get("PASSWORD"),
                host=values.get("SERVER"),
                port=str(values.get("PORT")),
                path=f"/{values.get('NAME') or ''}",
            )
        )

    class Config:
        """Configuration for the BaseSettings class."""

        env_prefix = "GWDB_"
        env_file = ".env"
        case_sensitive = True



"""Generate SQL queries to get Gene information."""

from typing import List, Optional, Tuple

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.core.mapping import AON_ID_TYPE_FOR_SPECIES
from geneweaver.db.utils import format_sql_fields, limit_and_offset
from psycopg.sql import SQL, Composed

GENE_FIELDS_MAP = {
    "ode_gene_id": "id",
    "ode_ref_id": "reference_id",
    "gdb_id": "gene_database",
    "sp_id": "species",
    "ode_pref": "preferred",
    "ode_date": "date",
}

GENE_FIELDS = format_sql_fields(GENE_FIELDS_MAP, query_table="gene")


def get(
    gene_id: Optional[int] = None,
    reference_id: Optional[str] = None,
    gene_database: Optional[GeneIdentifier] = None,
    species: Optional[Species] = None,
    preferred: Optional[bool] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> Tuple[Composed, dict]:
    """Create a psycopg query to Get genes from the database.

    :param gene_id: The id of the gene to get.
    :param reference_id: The reference id to search for.
    :param gene_database: The gene database to search for.
    :param species: The species to search for.
    :param preferred: Whether to search for preferred genes.
    :param limit: The limit of results to return.
    :param offset: The offset of results to return.

    :return:  A query (and params) that can be executed on a cursor.
    """
    params = {}
    query = SQL("SELECT") + SQL(",").join(GENE_FIELDS) + SQL("FROM gene")

    filtering = []

    if gene_id:
        filtering.append(SQL("ode_gene_id = %(gene_id)s"))
        params["gene_id"] = gene_id

    if reference_id:
        filtering.append(SQL("ode_ref_id = %(ref_id)s"))
        params["ref_id"] = reference_id

    if gene_database:
        filtering.append(SQL("gdb_id = %(gene_db_id)s"))
        params["gene_db_id"] = int(gene_database)

    if species:
        filtering.append(SQL("sp_id = %(species_id)s"))
        params["species_id"] = int(species)

    if preferred is not None:
        filtering.append(SQL("ode_pref = %(preferred)s"))
        params["preferred"] = preferred

    if len(filtering) > 0:
        query += SQL("WHERE") + SQL(" AND ").join(filtering)

    query = limit_and_offset(query, limit, offset).join(" ")

    return query, params


def mapping(
    source_ids: List[str],
    species: Species,
    target_gene_id_type: GeneIdentifier,
) -> Tuple[Composed, dict]:
    """Create a query to get a list of gene IDs in an alternate identifier type.

    :param source_ids: The list of gene IDs to map.
    :param target_gene_id_type: The gene identifier type to map to.
    :param species: The species of the identifiers.
    :return: A query (and params) that can be executed on a cursor.
    """
    query = (
        SQL(
            """
    WITH PrefTrueCheck AS
        (SELECT CASE
                WHEN EXISTS(SELECT 1
                            FROM extsrc.gene AS g1
                                   JOIN extsrc.gene AS g2
                                        ON g1.ode_gene_id = g2.ode_gene_id
                                        AND g1.ode_ref_id != g2.ode_ref_id
                                        AND g1.sp_id = g2.sp_id
                             WHERE g1.ode_ref_id = ANY(%(source_ids)s)
                               AND g2.gdb_id = %(target_gene_id_type)s
                               AND g2.sp_id = %(species_id)s
                               AND g2.ode_pref = True)
                  THEN True ELSE False
                  END AS PrefExists)
    """
        )
        + SQL(
            """
    SELECT g1.ode_ref_id AS original_ref_id,
           g2.ode_ref_id AS mapped_ref_id
    FROM extsrc.gene AS g1
             JOIN extsrc.gene AS g2
                  ON g1.ode_gene_id = g2.ode_gene_id
                  AND g1.ode_ref_id != g2.ode_ref_id
                  AND g1.sp_id = g2.sp_id,
         PrefTrueCheck
    WHERE g1.ode_ref_id = ANY (%(source_ids)s)
      AND g2.gdb_id = %(target_gene_id_type)s
      AND g2.sp_id = %(species_id)s
      AND (
        (PrefTrueCheck.PrefExists AND g2.ode_pref = True)
            OR
        (NOT PrefTrueCheck.PrefExists)
        );
    """
        )
    ).join(" ")

    params = {
        "source_ids": source_ids,
        "target_gene_id_type": int(target_gene_id_type),
        "species_id": int(species),
    }
    return query, params


def aon_mapping(
    source_ids: List[str],
    species: Species,
) -> Tuple[Composed, dict]:
    """Create a query to get a list of gene IDs the AON preferred identifier type.

    :param source_ids: The list of gene IDs to map.
    :param species: The species to map from.
    :return: A query (and params) that can be executed on a cursor.
    """
    target_gene_id_type = AON_ID_TYPE_FOR_SPECIES[species]
    return mapping(source_ids, species, target_gene_id_type)



"""Database functions for geneset values."""

from typing import List, Optional

from geneweaver.core.enum import GeneIdentifier
from geneweaver.core.schema.batch import GenesetValueInput
from geneweaver.db.exceptions import GeneweaverTypeError
from psycopg import Cursor


def format_geneset_values_for_file_insert(
    geneset_values: List[GenesetValueInput],
) -> str:
    """Format geneset values for insertion into the database.

    :param geneset_values: List of geneset values to insert.
    :return: A string to insert into the database.
    """
    return "\n".join(
        (
            f"{geneset_value.symbol}\t{geneset_value.value}"
            for geneset_value in geneset_values
        )
    )


def insert_file(
    cursor: Cursor, formatted_geneset_values: str, comments: str = ""
) -> int:
    """Insert a file of geneset values into the database.

    :param cursor: The database cursor.
    :param formatted_geneset_values: A string formatted version of the geneset values.
    See the `format_geneset_values_for_file_insert` for formatting requirements.
    :param comments: misc. comments about this file

    :raises GeneweaverTypeError: If the geneset values are not a string or are empty.

    :return: The file ID of the inserted file.
    """
    if (
        not isinstance(formatted_geneset_values, str)
        or len(formatted_geneset_values) == 0
    ):
        raise GeneweaverTypeError("Geneset values must be a nonzero length string.")

    cursor.execute(
        """
        INSERT INTO file
            (file_size, file_contents, file_comments, file_created)
        VALUES
            (%s, %s, %s, NOW())
        RETURNING file_id;
        """,
        (len(formatted_geneset_values), formatted_geneset_values, comments),
    )

    cursor.connection.commit()

    return cursor.fetchone()[0]


def get_file(cursor: Cursor, file_id: int) -> str:
    """Retrieve the file contents of a geneset.

    :param cursor: The database cursor.
    :param file_id: File ID associated with a geneset

    :return: A string formatted version of the geneset values. An empty string if the
    file does not exist.
    """
    cursor.execute(
        """
        SELECT file_contents FROM file WHERE file_id = %(file_id)s;
        """,
        {"file_id": file_id},
    )
    contents = cursor.fetchone()

    if not contents:
        return ""

    return contents[0]


def insert_geneset_value(
    cursor: Cursor,
    geneset_id: int,
    gene_id: int,
    value: str,
    name: str,
    within_threshold: bool,
) -> int:
    """Insert a geneset value into the database.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to insert the value into.
    :param gene_id: The gene ID (ode_gene_id) to insert the value into.
    :param value: The value associated with this gene.
    :param name: A gene name or symbol (typically an ode_ref_id).
    :param within_threshold: Whether the value is within the threshold.

    :return: The geneset value ID of the inserted value.
    """
    cursor.execute(
        """
        INSERT INTO geneset_value

            (gs_id, ode_gene_id, gsv_value, gsv_source_list,
            gsv_value_list, gsv_in_threshold, gsv_hits, gsv_date)

        VALUES

            (%s, %s, %s, %s, %s, %s, 0, NOW())

        RETURNING gs_id;
        """,
        (geneset_id, gene_id, value, [name], [float(value)], within_threshold),
    )

    cursor.connection.commit()

    return cursor.fetchone()[0]


def by_geneset_id(
    cursor: Cursor, geneset_id: int, identifier: Optional[GeneIdentifier] = None
) -> list:
    """Retrieve all geneset values associated with a geneset.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to retrieve values for.
    :param identifier: The gene identifier to return.

    :return: A list of geneset values associated with the geneset.
    """
    if identifier is not None:
        return by_geneset_id_and_identifier(cursor, geneset_id, identifier)
    else:
        return by_geneset_id_as_uploaded(cursor, geneset_id)


def by_geneset_id_as_uploaded(cursor: Cursor, geneset_id: int) -> list:
    """Retrieve all geneset values associated with a geneset.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to retrieve values for.

    :return: A list of geneset values associated with the geneset.
    """
    cursor.execute(
        """
        SELECT DISTINCT ON (gv.ode_gene_id) gv.*, g.ode_ref_id
        FROM        extsrc.geneset_value gv
        INNER JOIN  extsrc.gene g
        USING       (ode_gene_id)
        WHERE  gs_id = %(geneset_id)s;
        """,
        {"geneset_id": geneset_id},
    )
    return cursor.fetchall()


def by_geneset_id_and_identifier(
    cursor: Cursor, geneset_id: int, identifier: GeneIdentifier
) -> list:
    """Retrieve all geneset values associated with a geneset.

    NOTE: If you are mapping identifiers across species, you will need to use the
    `geneweaver.db.gene.get_homolog_ids_by_ode_id` function to get the homolog ids
    for the geneset values after processing the results of this function.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to retrieve values for.
    :param identifier: The gene identifier to use.

    :return: A list of geneset values associated with the geneset.
    """
    cursor.execute(
        """
            SELECT gsv.gs_id, gsv.ode_gene_id, gsv.gsv_value, gsv.gsv_hits,
                   gsv.gsv_source_list, gsv.gsv_value_list,
                   gsv.gsv_in_threshold, gsv.gsv_date, h.hom_id, gi.gene_rank,
                   gsv.ode_ref_id, gsv.gdb_id

            --
            -- Use a subquery here so we can prevent duplicate gene identifiers
            -- of the same type from being returned (the DISTINCT ON section)
            -- otherwise when we try to change identifier types from the view
            -- GS page, duplicate entries screw things up
            --
            FROM (
                SELECT DISTINCT ON (g.ode_gene_id, g.gdb_id)
                        gsv.*, g.ode_ref_id, g.gdb_id, g.ode_pref
                FROM    geneset_value as gsv, gene as g
                WHERE   gsv.gs_id = %(geneset_id)s AND
                        g.ode_gene_id = gsv.ode_gene_id AND
                        g.gdb_id = (SELECT COALESCE (
                            (SELECT gdb_id
                             FROM   gene AS g2
                             WHERE g2.ode_gene_id = gsv.ode_gene_id AND
                                   g2.gdb_id = %(gdb_id)s
                             LIMIT 1),
                            (SELECT gdb_id
                             FROM   gene AS g2
                             WHERE g2.ode_gene_id = gsv.ode_gene_id AND
                                   g2.gdb_id = 7
                             LIMIT 1)
                        )) AND

                        --
                        -- When viewing symbols, always pick the preferred gene symbol
                        --
                        CASE
                            WHEN g.gdb_id = 7 THEN g.ode_pref = 't'
                            ELSE true
                        END
            ) gsv

            --
            -- gene_info necessary for the priority scores
            --
            INNER JOIN  gene_info AS gi
            ON          gsv.ode_gene_id = gi.ode_gene_id

            --
            -- Have to use a left outer join because some genes may not have homologs
            --
            LEFT OUTER JOIN homology AS h
            ON          gsv.ode_gene_id = h.ode_gene_id

            WHERE h.hom_source_name = 'Homologene' OR
                  -- In case the gene doesn't have any homologs
                  h.hom_source_name IS NULL
        """,
        {"geneset_id": geneset_id, "gdb_id": int(identifier)},
    )
    return cursor.fetchall()



"""Database code for interacting with Publication table."""

from typing import Iterable, List, Optional

from geneweaver.core.schema.publication import PublicationInfo
from geneweaver.db.query import publication as publication_query
from psycopg import Cursor, rows


def by_pubmed_id(cursor: Cursor, pubmed_id: int) -> Optional[rows.Row]:
    """Get a publication by PubMed ID.

    :param cursor: The database cursor.
    :param pubmed_id: The PubMed ID to search for.

    :return: optional row using `.fetchone()`
    """
    cursor.execute(*publication_query.by_pubmed_id(pubmed_id))
    return cursor.fetchone()


def by_pubmed_ids(cursor: Cursor, pubmed_ids: Iterable[int]) -> List[rows.Row]:
    """Get publications by a list of PubMed IDs.

    :param cursor: The database cursor.
    :param pubmed_ids: The PubMed IDs to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(*publication_query.by_pubmed_ids(pubmed_ids))
    return cursor.fetchall()


def by_id(cursor: Cursor, pub_id: int) -> Optional[rows.Row]:
    """Get a publication by ID.

    :param cursor: The database cursor.
    :param pub_id: The publication ID (geneweaver internal) to search for.

    :return: optional row using `.fetchone()`
    """
    cursor.execute(*publication_query.by_id(pub_id))
    return cursor.fetchone()


def by_geneset_id(cursor: Cursor, geneset_id: int) -> Optional[rows.Row]:
    """Get a publication by geneset ID.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to search for.

    :return: optional row using `.fetchone()`
    """
    cursor.execute(*publication_query.by_geneset_id(geneset_id))
    return cursor.fetchone()


def add(cursor: Cursor, publication: PublicationInfo) -> Optional[rows.Row]:
    """Add a publication to the database.

    :param cursor: The database cursor.
    :param publication: The publication to add.

    :return: optional row using `.fetchone()`
    """
    cursor.execute(*publication_query.add(**publication.dict()))
    return cursor.fetchone()



"""Generate SQL queries to get Species information."""

from typing import Optional, Tuple

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.db.utils import format_sql_fields
from psycopg.sql import SQL, Composed

SPECIES_FIELD_MAP = {
    "sp_id": "id",
    "sp_name": "name",
    "sp_taxid": "taxonomic_id",
    "sp_ref_gdb_id": "reference_gene_identifier",
}


SPECIES_FIELDS = format_sql_fields(SPECIES_FIELD_MAP, query_table="species")


def get(
    taxonomic_id: Optional[int] = None,
    reference_gene_db_id: Optional[GeneIdentifier] = None,
    species: Optional[Species] = None,
) -> Tuple[Composed, dict]:
    """Create a psycopg query to get species information from the database.

    :param taxonomic_id: The taxonomic id.
    :param reference_gene_db_id: The reference gene database id.
    :param species: The species id.

    :return: A query (and params) that can be executed on a cursor.
    """
    params = {}
    query = SQL("SELECT") + SQL(",").join(SPECIES_FIELDS) + SQL("FROM species")

    filters = []

    if species:
        filters.append(SQL("sp_id = %(species_id)s"))
        params["species_id"] = int(species)

    if taxonomic_id:
        filters.append(SQL("sp_taxid = %(taxonomic_id)s"))
        params["taxonomic_id"] = taxonomic_id

    if reference_gene_db_id:
        filters.append(SQL("sp_ref_gdb_id = %(reference_gene_db_id)s"))
        params["reference_gene_db_id"] = int(reference_gene_db_id)

    if len(filters) > 0:
        query += SQL("WHERE") + SQL("AND").join(filters)

    query = query.join(" ")

    return query, params



"""Exceptions for the geneweaver db module."""

from geneweaver.core.exc import GeneweaverError


class GeneweaverTypeError(GeneweaverError):
    """Exception raised when a type error occurs."""

    pass


class GeneweaverValueError(GeneweaverError):
    """Exception raised when a value error occurs."""

    pass


class GeneweaverDoesNotExistError(GeneweaverError):
    """Exception raised when an object does not exist."""

    pass



"""Database interaction code relating to Gene IDs."""

from typing import List, Optional

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.db.query import gene as gene_query
from psycopg import AsyncCursor, rows


async def get(
    cursor: AsyncCursor,
    reference_id: Optional[str] = None,
    gene_database: Optional[GeneIdentifier] = None,
    species: Optional[Species] = None,
    preferred: Optional[bool] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List:
    """Get genes from the database.

    :param cursor: An async database cursor.
    :param reference_id: The reference id to search for.
    :param gene_database: The gene database to search for.
    :param species: The species to search for.
    :param preferred: Whether to search for preferred genes.
    :param limit: The limit of results to return.
    :param offset: The offset of results to return.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *gene_query.get(
            reference_id=reference_id,
            gene_database=gene_database,
            species=species,
            preferred=preferred,
            limit=limit,
            offset=offset,
        )
    )

    return await cursor.fetchall()


async def get_preferred(
    cursor: AsyncCursor,
    gene_id: int,
) -> Optional[rows.Row]:
    """Get the preferred gene from the database for a given ode_id.

    :param cursor: An async database cursor.
    :param gene_id: The id of the gene to get.

    :return: The preferred gene using `.fetchone()`
    """
    await cursor.execute(
        *gene_query.get(
            gene_id=gene_id,
            preferred=True,
        )
    )

    return await cursor.fetchone()


async def mapping(
    cursor: AsyncCursor,
    source_ids: List[str],
    species: Species,
    target_gene_id_type: GeneIdentifier,
) -> List:
    """Get gene mappings from the database.

    This method works _within_ a species.

    :param cursor: An async database cursor.
    :param source_ids: The list of gene ids to get mappings for.
    :param species: The species of the identifiers.
    :param target_gene_id_type: The gene id type to return.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *gene_query.mapping(
            source_ids=source_ids,
            species=species,
            target_gene_id_type=target_gene_id_type,
        )
    )

    return await cursor.fetchall()


async def aon_mapping(
    cursor: AsyncCursor,
    source_ids: List[str],
    species: Species,
) -> List:
    """Get gene mappings in the default identifier type for that species in AON.

    This method works _within_ a species.

    :param cursor: An async database cursor.
    :param source_ids: The list of gene ids to get mappings for.
    :param species: The species of the identifiers.

    :return: list of results using `.fetchall()`
    """
    await cursor.execute(
        *gene_query.aon_mapping(
            source_ids=source_ids,
            species=species,
        )
    )

    return await cursor.fetchall()



"""Module for all geneset query generation functions."""

# ruff: noqa: F401

from geneweaver.db.query.geneset.const import (
    GENESET_FIELDS,
    GENESET_TSVECTOR,
    PUB_FIELDS,
)
from geneweaver.db.query.geneset.read import by_project_id, get
from geneweaver.db.query.geneset.write import (
    add,
    add_geneset_file,
    render_and_add_geneset_file,
)



"""SQL query generation code for reading genesets."""

from typing import Optional, Tuple

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.db.query.geneset.utils import (
    GenesetTierOrTiers,
    format_select_query,
    is_readable,
    restrict_tier,
    search,
)
from geneweaver.db.query.utils import construct_filters
from geneweaver.db.utils import limit_and_offset
from psycopg.sql import SQL, Composed


def get(
    gs_id: Optional[int] = None,
    owner_id: Optional[int] = None,
    curation_tier: Optional[GenesetTierOrTiers] = None,
    species: Optional[Species] = None,
    name: Optional[str] = None,
    abbreviation: Optional[str] = None,
    publication_id: Optional[int] = None,
    pubmed_id: Optional[int] = None,
    gene_id_type: Optional[GeneIdentifier] = None,
    search_text: Optional[str] = None,
    status: Optional[str] = "normal",
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> Tuple[Composed, dict]:
    """Get genesets.

    :param gs_id: Show only results with this geneset ID.
    :param owner_id: Show only results owned by this user ID.
    :param curation_tier: Show only results of this curation tier.
    :param species: Show only results associated with this species.
    :param name: Show only results with this name.
    :param abbreviation: Show only results with this abbreviation.
    :param publication_id: Show only results with this publication ID (internal).
    :param pubmed_id: Show only results with this PubMed ID.
    :param gene_id_type: Show only results with this gene ID type.
    :param search_text: Return genesets that match this search text (using PostgreSQL
                        full-text search).
    :param status: Show only results with this status. Default is "normal".
    :param limit: Limit the number of results.
    :param offset: Offset the results.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.
    """
    params = {}
    filtering = []
    query = format_select_query(
        with_publication_info=with_publication_info,
        with_publication_join=pubmed_id is not None,
    )

    filtering, params = is_readable(filtering, params, is_readable_by)
    filtering, params = search(filtering, params, search_text)
    filtering, params = restrict_tier(filtering, params, curation_tier)

    filtering, params = construct_filters(
        filtering,
        params,
        {
            "gs_id": gs_id,
            "usr_id": owner_id,
            "sp_id": int(species) if species is not None else None,
            "gs_name": name,
            "gs_abbreviation": abbreviation,
            "pub_id": publication_id,
            "publication.pubmed_id": pubmed_id,
            "gs_gene_id_type": int(gene_id_type) if gene_id_type is not None else None,
            "gs_status": status,
        },
    )

    if len(filtering) > 0:
        query += SQL("WHERE") + SQL("AND").join(filtering)

    query = limit_and_offset(query, limit, offset).join(" ")

    return query, params


def shared_with_user(
    user_id: int,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    with_publication_info: bool = True,
) -> Tuple[Composed, dict]:
    """Get Genesets that are shared with a user.

    NOTE: NOT IMPLEMENTED
    """
    raise NotImplementedError()


def by_project_id(
    project_id: int,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> Tuple[Composed, dict]:
    """Create a psycopg query to get genesets by project ID.

    :param project_id: The project ID to search for.
    :param limit: The limit of results to return.
    :param offset: The offset of results to return.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: A query (and params) that can be executed on a cursor.
    """
    query = (
        format_select_query(with_publication_info=with_publication_info)
        + SQL("JOIN project_geneset ON geneset.gs_id = project_geneset.gs_id")
        + SQL("INNER JOIN production.project2geneset")
        + SQL("ON geneset.gs_id=project2geneset.gs_id")
        + SQL("WHERE project2geneset.pj_id=%(project_id)s")
    )
    params = {"project_id": project_id}
    filtering, params = is_readable([], params, is_readable_by)

    if len(filtering) > 0:
        query += SQL("WHERE") + SQL("AND").join(filtering)
    query = limit_and_offset(query, limit, offset).join(" ")

    return query, params



"""Database interaction code relating to Gene IDs."""

from typing import Iterable, List, Optional

from geneweaver.core.enum import GeneIdentifier, Species
from geneweaver.db.query import gene as gene_query
from psycopg import Cursor, rows
from psycopg.sql import SQL


def get(
    cursor: Cursor,
    reference_id: Optional[str] = None,
    gene_database: Optional[GeneIdentifier] = None,
    species: Optional[Species] = None,
    preferred: Optional[bool] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> List:
    """Get genes from the database.

    :param cursor: The database cursor.
    :param reference_id: The reference id to search for.
    :param gene_database: The gene database to search for.
    :param species: The species to search for.
    :param preferred: Whether to search for preferred genes.
    :param limit: The limit of results to return.
    :param offset: The offset of results to return.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *gene_query.get(
            reference_id=reference_id,
            gene_database=gene_database,
            species=species,
            preferred=preferred,
            limit=limit,
            offset=offset,
        )
    )

    return cursor.fetchall()


def get_preferred(
    cursor: Cursor,
    gene_id: int,
) -> Optional[rows.Row]:
    """Get the preferred gene from the database for a given ode_id.

    :param cursor: The database cursor.
    :param gene_id: The id of the gene to get.

    :return: The preferred gene using `.fetchone()`
    """
    cursor.execute(
        *gene_query.get(
            gene_id=gene_id,
            preferred=True,
        )
    )

    return cursor.fetchone()


def mapping(
    cursor: Cursor,
    source_ids: List[str],
    species: Species,
    target_gene_id_type: GeneIdentifier,
) -> List:
    """Get gene mappings from the database.

    This method works _within_ a species.

    :param cursor: An async database cursor.
    :param source_ids: The list of gene ids to get mappings for.
    :param species: The species of the identifiers.
    :param target_gene_id_type: The gene id type to return.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *gene_query.mapping(
            source_ids=source_ids,
            species=species,
            target_gene_id_type=target_gene_id_type,
        )
    )

    return cursor.fetchall()


def aon_mapping(
    cursor: Cursor,
    source_ids: List[str],
    species: Species,
) -> List:
    """Get gene mappings in the default identifier type for that species in AON.

    This method works _within_ a species.

    :param cursor: An async database cursor.
    :param source_ids: The list of gene ids to get mappings for.
    :param species: The species of the identifiers.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *gene_query.aon_mapping(
            source_ids=source_ids,
            species=species,
        )
    )

    return cursor.fetchall()


# --------------------------------------------------------------------------------------
# NOTE: The following functions are not yet migrated to the new query/execute pattern.
# --------------------------------------------------------------------------------------


def id_types(cursor: Cursor, species: Optional[Species] = None) -> List:
    """Get all the Gene ID types from the database.

    :param cursor: The database cursor.
    :param species: Limit to additional species other than mouse

    :return: list of results using `.fetchall()`
    """
    if species is None:
        cursor.execute("""SELECT * FROM odestatic.genedb ORDER BY gdb_id;""")
    else:
        cursor.execute(
            """SELECT * FROM odestatic.genedb
            WHERE sp_id=0 OR sp_id=%(sp_id)s
            ORDER BY gdb_id;""",
            {"sp_id": int(species)},
        )

    return cursor.fetchall()


# This SQL might be re-used by other functions, so it's defined here.
# Specifically, postgres has a row_to_json function that is used extensively for the
# GW2 API, and it's possible that we might want to use it here as well.
INFO_BY_GENE_ID_SQL = "SELECT * FROM extsrc.gene_info WHERE ode_gene_id = %(gene_id)s;"


def info_by_gene_id(cursor: Cursor, gene_id: int) -> List:
    """Get Gene ID type info by gene id.

    :param cursor: The database cursor.
    :param gene_id: The gene id to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(INFO_BY_GENE_ID_SQL, {"gene_id": gene_id})
    return cursor.fetchall()


def gene_database_by_id(cursor: Cursor, genedb_id: GeneIdentifier) -> List:
    """Get all gene database info by gene database id.

    :param cursor: The database cursor.
    :param genedb_id: The gene database id to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT * FROM odestatic.genedb WHERE gdb_id = %(gdb_id)s;""",
        {"gdb_id": int(genedb_id)},
    )
    return cursor.fetchall()


def gene_database_id(cursor: Cursor, identifier: GeneIdentifier) -> List:
    """Get all gene database info by gene database name.

    :param cursor: The database cursor.
    :param identifier: The gene database name to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT gdb_id FROM odestatic.genedb WHERE gdb_name = %(gdb_name)s;""",
        {"gdb_name": str(identifier)},
    )
    return cursor.fetchall()


def symbols_by_geneset_id(cursor: Cursor, geneset_id: int) -> List:
    """Get all gene symbols associated with a specific geneset id.

    :param cursor: The database cursor.
    :param geneset_id: The geneset id to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """
        SELECT g.ode_ref_id
        FROM extsrc.gene g, extsrc.geneset_value gv
        WHERE gv.gs_id= %(geneset_id)s
        AND gv.ode_gene_id=g.ode_gene_id
        AND g.gdb_id=7
        AND ode_pref='t';
        """,
        {"geneset_id": geneset_id},
    )
    return cursor.fetchall()


def symbols_by_project_id(cursor: Cursor, project_id: int) -> List:
    """Get all gene symbols associated with a specific project id.

    :param cursor: The database cursor.
    :param project_id: The project id to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """
        SELECT g.ode_ref_id
        FROM extsrc.gene g, extsrc.geneset_value gv
        WHERE (gv.gs_id IN
                (SELECT gs_id AS geneSetId
                FROM production.project2geneset
                WHERE pj_id = %(project_id)s))
        AND gv.ode_gene_id=g.ode_gene_id AND g.gdb_id=7 AND ode_pref='t';
    """,
        {"project_id": project_id},
    )
    return cursor.fetchall()


def get_homolog_ids_by_ode_id(
    cursor: Cursor, ode_gene_ids: Iterable[str], identifier: GeneIdentifier
) -> List:
    """Get all homolog ids associated with a specific gene id.

    :param cursor: The database cursor.
    :param ode_gene_ids: The gene ids to search for.
    :param identifier: The identifier to return genes in.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """
        SELECT DISTINCT     h.ode_gene_id, g.ode_ref_id
        FROM                homology AS h
        INNER JOIN          homology AS h2
        ON                  h.hom_id = h2.hom_id
        INNER JOIN          gene AS g
        ON                  g.ode_gene_id = h2.ode_gene_id
        WHERE               h.ode_gene_id = ANY(%(ode_gene_ids)s) AND
                            g.gdb_id = %(genedb_id)s;
        """,
        {"ode_gene_ids": list(ode_gene_ids), "genedb_id": int(identifier)},
    )
    return cursor.fetchall()


def get_homolog_ids(
    cursor: Cursor,
    source_ids: Iterable[str],
    result_identifier: GeneIdentifier,
    source_identifier: Optional[GeneIdentifier] = None,
    result_species: Optional[Species] = None,
    source_species: Optional[Species] = None,
    only_preferred_ids: bool = True,
) -> list:
    """Get homologous GeneIDs for a list of source IDs.

    :param cursor: The database cursor.
    :param source_ids: The gene ids to search for.
    :param result_identifier: The identifier to return genes in.
    :param source_identifier: The identifier to search for genes in (optional).
    :param result_species: The species to return genes in (optional, but recommended).
    :param source_species: The species to search for genes in (optional, but
                           recommended).
    :param only_preferred_ids: Whether to return only genes with  preferred identifiers.
    """
    base_query = SQL(
        """
    SELECT DISTINCT source_gene.ode_ref_id AS source_ref_id,
                    source_gene.sp_id AS source_sp_id,
                    result_gene.ode_ref_id AS result_ref_id,
                    result_gene.sp_id AS result_sp_id
        FROM homology AS source_homology
            INNER JOIN homology AS result_homology
                ON source_homology.hom_id = result_homology.hom_id
            INNER JOIN gene AS result_gene
                ON result_gene.ode_gene_id = result_homology.ode_gene_id
            INNER JOIN gene AS source_gene
                ON source_gene.ode_gene_id = source_homology.ode_gene_id
        WHERE source_gene.ode_ref_id = ANY(%(source_ids)s)
            AND result_gene.gdb_id = %(result_genedb_id)s
            AND result_gene.ode_pref = %(ode_pref)s
    """
    )

    params = {
        "source_ids": list(source_ids),
        "result_genedb_id": int(result_identifier),
        "ode_pref": "f" if only_preferred_ids is False else "t",
    }

    if source_identifier is not None:
        base_query += SQL("AND source_gene.gdb_id = %(source_genedb_id)s")
        params["source_genedb_id"] = int(source_identifier)

    if result_species is not None:
        base_query += SQL("AND result_gene.sp_id = %(result_sp_id)s")
        params["result_sp_id"] = int(result_species)

    if source_species is not None:
        base_query += SQL("AND source_gene.sp_id = %(source_sp_id)s")
        params["source_sp_id"] = int(source_species)

    cursor.execute(base_query, params)
    return cursor.fetchall()



"""Utilities for the GeneWeaver database functions."""

# ruff: noqa: ANN001, ANN002, ANN003, ANN201, ANN202
import functools
from typing import List, Optional

from geneweaver.db.exceptions import GeneweaverDoesNotExistError, GeneweaverValueError
from psycopg import sql
from psycopg.rows import Row


def unpack_one_item_fetchall_results(results: List[Row]) -> List:
    """Unpack a single column from multiple rows of results.

    :param results: The results from a fetchall call.

    :raises GeneweaverDoesNotExistError: If the result is empty.
    :raises GeneweaverTooManyResultsError: If the result has more than one row.

    :return: The single row from the results.
    """
    if len(results) == 0:
        raise GeneweaverDoesNotExistError("No results found.")

    if len(results[0]) > 1:
        raise GeneweaverValueError("Too many results to unpack.")

    return [t[0] for t in results]


def temp_override_row_factory(row_factory):
    """Temporarily override the row factory for a function.

    This is useful for functions that use a cursor as an argument, but which process
    cursor results internally and therefore need a specific row factory to be set.

    :param row_factory: The row factory to use.

    :return: The function wrapper.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Determine cursor based on positional or keyword argument
            if args:
                cursor = args[0]
            else:
                cursor = kwargs.get("cursor")

            if cursor is None:
                raise ValueError(
                    "Cursor must be provided either as a "
                    "positional or keyword argument."
                )

            original_row_factory = cursor.row_factory
            cursor.row_factory = row_factory

            try:
                result = func(*args, **kwargs)
            finally:
                cursor.row_factory = original_row_factory

            return result

        return wrapper

    return decorator


def format_sql_fields(
    fields_map: dict,
    query_table: Optional[str] = None,
    resp_prefix: Optional[str] = None,
) -> List[sql.Composed]:
    """Format SQL fields for a query."""
    select_prefix = query_table or ""
    resp_prefix = "" if resp_prefix is None else f"{resp_prefix}_"

    return [
        sql.SQL("{select_prefix}.{source_field} AS {to_field}").format(
            select_prefix=sql.Identifier(select_prefix),
            source_field=sql.Identifier(source),
            to_field=sql.Identifier(f"{resp_prefix}{mapping}"),
        )
        for source, mapping in fields_map.items()
    ]


def limit_and_offset(
    query: sql.Composed,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> sql.Composed:
    """Format SQL limit and offset clauses."""
    if limit is not None:
        query = query + sql.SQL("LIMIT {limit}").format(limit=limit)
    if offset is not None:
        query = query + sql.SQL("OFFSET {offset}").format(offset=offset)

    return query



"""Query functions related to thresholding and score type."""

from typing import Tuple

from geneweaver.core.schema.score import GenesetScoreType
from psycopg.sql import SQL, Composed


def set_geneset_threshold(
    geneset_id: int,
    geneset_score_type: GenesetScoreType,
) -> Tuple[Composed, dict]:
    """Update the threshold of a geneset and its values.

    :param geneset_id: The ID of the geneset to update.
    :param geneset_score_type: An instance of GenesetScoreType, representing the new
    threshold to set.
    :return:  A query (and params) that can be executed on a cursor.
    """
    params = {
        "geneset_id": geneset_id,
        "score_type": int(geneset_score_type.score_type),
        "threshold_str": geneset_score_type.threshold_as_db_string(),
        "threshold_high": geneset_score_type.threshold,
    }

    query = SQL(
        """
        UPDATE geneset
        SET gs_threshold_type = %(score_type)s,
            gs_threshold = %(threshold_str)s
        WHERE id = %(geneset_id)s;
        UPDATE geneset_value
            SET gsv_in_threshold = CASE
        """
    )
    if geneset_score_type.threshold_low is not None:
        query = query + SQL(
            """
            WHEN gsv_value BETWEEN %(threshold_low)s AND %(threshold_high)s THEN TRUE
            """
        )
        params["threshold_low"] = geneset_score_type.threshold_low

        if geneset_score_type.threshold_low > geneset_score_type.threshold:
            raise ValueError(
                "geneset_score_type.threshold must be larger than "
                "geneset_score_type.threshold_low"
            )

    else:
        query = query + SQL(
            """
            WHEN gsv_value < %(threshold_high)s THEN TRUE
            """
        )

    query = query + SQL(
        """
                ELSE FALSE
            END
        WHERE gs_id = %(geneset_id)s;
        """
    )

    query = query.join(" ")

    return query, params



"""Constants for use in geneset queries."""

from geneweaver.db.query.publication import PUB_FIELD_MAP
from geneweaver.db.utils import format_sql_fields
from psycopg.sql import SQL, Identifier

GENESET_FIELDS_MAP = {
    "gs_id": "id",
    "usr_id": "user_id",
    "file_id": "file_id",
    "cur_id": "curation_id",
    "sp_id": "species_id",
    "gs_name": "name",
    "gs_abbreviation": "abbreviation",
    "pub_id": "publication_id",
    "gs_description": "description",
    "gs_count": "count",
    "gs_threshold_type": "score_type",
    "gs_threshold": "threshold",
    "gs_status": "status",
    "gs_gene_id_type": "gene_id_type",
    "gs_attribution": "attribution",
    "gs_created": "created",
    "gs_updated": "updated",
}

GENESET_FIELDS = format_sql_fields(GENESET_FIELDS_MAP, query_table="geneset")
PUB_FIELDS = format_sql_fields(
    PUB_FIELD_MAP, query_table="publication", resp_prefix="publication"
)
GENESET_TSVECTOR = (Identifier("geneset") + Identifier("gs_tsvector")).join(".")

COPY_GENESET_VALUES = SQL(
    """
                COPY geneset_value
                (gs_id,
                ode_gene_id,
                gsv_value,
                gsv_source_list,
                gsv_value_list,
                gsv_in_threshold,
                gsv_hits,
                gsv_date)
                FROM STDIN
                """
)



"""Geneset database functions."""

from typing import List, Optional

from geneweaver.core.enum import GeneIdentifier, GenesetTier, Species
from geneweaver.db.query import geneset as geneset_query
from geneweaver.db.utils import temp_override_row_factory
from psycopg import Cursor, rows
from psycopg.rows import Row


def get(
    cursor: Cursor,
    is_readable_by: Optional[int] = None,
    gs_id: Optional[int] = None,
    owner_id: Optional[int] = None,
    curation_tier: Optional[GenesetTier] = None,
    species: Optional[Species] = None,
    name: Optional[str] = None,
    abbreviation: Optional[str] = None,
    publication_id: Optional[int] = None,
    pubmed_id: Optional[int] = None,
    gene_id_type: Optional[GeneIdentifier] = None,
    search_text: Optional[str] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get genesets from the database.

    :param cursor: An async database cursor.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param gs_id: Show only results with this geneset ID.
    :param owner_id: Show only results owned by this user ID.
    :param curation_tier: Show only results of this curation tier.
    :param species: Show only results associated with this species.
    :param name: Show only results with this name.
    :param abbreviation: Show only results with this abbreviation.
    :param publication_id: Show only results with this publication ID (internal).
    :param pubmed_id: Show only results with this PubMed ID.
    :param gene_id_type: Show only results with this gene ID type.
    :param search_text: Return genesets that match this search text (using PostgreSQL
                        full-text search).
    :param limit: Limit the number of results.
    :param offset: Offset the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *geneset_query.get(
            is_readable_by=is_readable_by,
            gs_id=gs_id,
            owner_id=owner_id,
            curation_tier=curation_tier,
            species=species,
            name=name,
            abbreviation=abbreviation,
            publication_id=publication_id,
            pubmed_id=pubmed_id,
            gene_id_type=gene_id_type,
            search_text=search_text,
            limit=limit,
            offset=offset,
            with_publication_info=with_publication_info,
        )
    )

    return cursor.fetchall()


def by_id(
    cursor: Cursor,
    gs_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> Optional[Row]:
    """Get a geneset by its ID.

    :param cursor: A database cursor.
    :param gs_id: The geneset ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.
    """
    cursor.execute(
        *geneset_query.get(
            gs_id=gs_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )

    return cursor.fetchone()


def by_owner_id(
    cursor: Cursor,
    owner_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get genesets by owner ID.

    :param cursor: A database cursor.
    :param owner_id: The owner ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *geneset_query.get(
            owner_id=owner_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )
    return cursor.fetchall()


def by_project_id(
    cursor: Cursor,
    project_id: int,
    is_readable_by: Optional[int] = None,
    with_publication_info: bool = True,
) -> List[Row]:
    """Get a list of genesets by their membership in a project.

    :param cursor: An async database cursor.
    :param project_id: The project ID to search for.
    :param is_readable_by: A user ID to check if the user can read the results.
    :param with_publication_info: Include publication info in the return.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        *geneset_query.by_project_id(
            project_id=project_id,
            is_readable_by=is_readable_by,
            with_publication_info=with_publication_info,
        )
    )
    return cursor.fetchall()


# --------------------------------------------------------------------------------------
# The following functions are not yet implemented in the aio/concurrent paradigm.
# They are subject to change and may be deprecated in the future.


@temp_override_row_factory(rows.tuple_row)
def is_readable(cursor: Cursor, user_id: int, geneset_id: int) -> bool:
    """Check if a geneset is readable by a user.

    :param cursor: The database cursor.
    :param user_id: The user id (internal) to check.
    :param geneset_id: The geneset id to check.

    :return: True if the geneset is readable by the user, False otherwise.
    """
    cursor.execute(
        """
        SELECT production.geneset_is_readable2(%(user_id)s, %(geneset_id)s);
        """,
        {"user_id": user_id, "geneset_id": geneset_id},
    )
    return cursor.fetchone()[0] is True


@temp_override_row_factory(rows.tuple_row)
def user_is_owner(cursor: Cursor, user_id: int, geneset_id: int) -> bool:
    """Check if a user is the owner of a geneset.

    :param cursor: The database cursor.
    :param user_id: The user id (internal) to check.
    :param geneset_id: The geneset id to check.

    :return: True if the user is the owner of the geneset, False otherwise.
    """
    cursor.execute(
        """
        SELECT COUNT(gs_id) FROM geneset
        WHERE usr_id=%(user_id)s AND gs_id=%(geneset_id)s;
        """,
        {"user_id": user_id, "geneset_id": geneset_id},
    )
    result = cursor.fetchone()[0]
    return result == 1 and not isinstance(result, bool)


@temp_override_row_factory(rows.tuple_row)
def update_date(cursor: Cursor, geneset_id: int) -> str:
    """Update the date of a geneset.

    :param cursor: The database cursor.
    :param geneset_id: The geneset id to update.

    :return: The updated date.
    """
    cursor.execute(
        """
        UPDATE geneset SET gs_updated = NOW()
        WHERE gs_id = %(geneset_id)s
        RETURNING gs_updated
        """,
        {"geneset_id": geneset_id},
    )
    cursor.connection.commit()
    return cursor.fetchone()[0]


@temp_override_row_factory(rows.tuple_row)
def tier(cursor: Cursor, geneset_id: int) -> Optional[int]:
    """Get the tier of a geneset.

    :param cursor: The database cursor.
    :param geneset_id: The geneset id to get the tier of.

    :return: The tier of the geneset, or None if the geneset does not have a tier,
    or does not exist.
    """
    cursor.execute(
        """
        SELECT cur_id FROM geneset WHERE gs_id = %(geneset_id)s;
        """,
        {"geneset_id": geneset_id},
    )
    result = cursor.fetchone()

    if not result:
        return None

    return result[0]


def homology_ids(cursor: Cursor, geneset_id: int) -> List:
    """Get all homology_ids that are associated with a geneset ID.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """
        SELECT DISTINCT hom_id FROM extsrc.homology h
        INNER JOIN extsrc.geneset_value gsv
            ON h.ode_gene_id=gsv.ode_gene_id
        INNER JOIN production.geneset g
            ON gsv.gs_id=g.gs_id
        WHERE gs_status not like 'de%%' AND g.gs_id=%(geneset_id)s""",
        {"geneset_id": geneset_id},
    )
    return cursor.fetchall()


# TODO: reimplement `get_all_geneset_values`


@temp_override_row_factory(rows.tuple_row)
def num_genes(cursor: Cursor, geneset_id: int) -> int:
    """Get the number of genes associated with a geneset ID.

    :param cursor: The database cursor.
    :param geneset_id: The geneset ID to search for.

    :return: The number of genes associated with the geneset ID.
    """
    cursor.execute(
        """
        SELECT COUNT(*) FROM extsrc.geneset_value WHERE gs_id = %(geneset_id)s;
        """,
        {"geneset_id": geneset_id},
    )
    return cursor.fetchone()[0]



"""Convenience code for interacting with the database."""

from contextlib import asynccontextmanager, contextmanager

import psycopg
from geneweaver.db.core.settings import settings


@contextmanager
def cursor() -> psycopg.Cursor:
    """Get a cursor to the database."""
    with psycopg.connect(settings.URI) as connection:
        with connection.cursor() as _cursor:
            yield _cursor


@asynccontextmanager
async def async_cursor() -> psycopg.AsyncCursor:
    """Get an async cursor to the database."""
    async with await psycopg.AsyncConnection.connect(settings.URI) as connection:
        async with connection.cursor() as _cursor:
            yield _cursor


def make_connection() -> psycopg.Connection:
    """Make a connection to the database.

    Don't forget to close it!
    """
    return psycopg.connect(settings.URI)



"""SQL query generation code for writing genesets."""

from typing import List, Tuple

from geneweaver.core.enum import GeneIdentifier, GenesetTier, Species
from geneweaver.core.schema.gene import GeneValue
from psycopg.sql import SQL, Composed


def add(
    user_id: int,
    file_id: int,
    name: str,
    abbreviation: str,
    publication_id: int,
    tier: GenesetTier,
    description: str,
    species: Species,
    count: int,
    threshold_type: str,
    threshold: float,
    gene_id_type: GeneIdentifier,
    attribution: str,
) -> Tuple[Composed, dict]:
    """Add a geneset to the database.

    :param user_id: The user ID of the geneset owner.
    :param file_id: The file ID of the geneset file.
    :param name: The name of the geneset.
    :param abbreviation: The abbreviation of the geneset.
    :param publication_id: The publication ID of the geneset.
    :param tier: The curation tier of the geneset.
    :param description: The description of the geneset.
    :param species: The species of the geneset.
    :param count: The count of the geneset.
    :param threshold_type: The threshold type of the geneset.
    :param threshold: The threshold of the geneset.
    :param gene_id_type: The gene ID type of the geneset.
    :param attribution: The attribution of the geneset.

    :return: A query (and params) that can be executed on a cursor.
    """
    query_cols = SQL(
        """
        (usr_id, file_id, gs_name, gs_abbreviation, pub_id, cur_id,
        gs_description, sp_id, gs_count, gs_threshold_type,
        gs_threshold, gs_groups, gs_gene_id_type, gs_created,
        gs_attribution)
    """
    )
    query_vals = SQL(
        """
        VALUES
        (%(usr_id)s, %(file_id)s, %(gs_name)s, %(gs_abbreviation)s,
        %(pub_id)s, %(cur_id)s, %(gs_description)s, %(sp_id)s,
        %(gs_count)s, %(gs_threshold_type)s, %(gs_threshold)s,
        %(gs_groups)s, %(gs_gene_id_type)s, %(gs_created)s,
        %(gs_attribution)s
    """
    )

    query = (
        SQL("INSERT INTO geneset") + query_cols + query_vals + SQL("RETURNING gs_id")
    ).join(" ")

    params = {
        "usr_id": user_id,
        "file_id": file_id,
        "gs_name": name,
        "gs_abbreviation": abbreviation,
        "pub_id": publication_id,
        "cur_id": int(tier),
        "gs_description": description,
        "sp_id": int(species),
        "gs_count": count,
        "gs_threshold_type": threshold_type,
        "gs_threshold": threshold,
        "gs_groups": "",
        "gs_gene_id_type": int(gene_id_type),
        "gs_created": "NOW()",
        "gs_attribution": attribution,
    }

    return query, params


def render_and_add_geneset_file(
    values: List[GeneValue],
    comments: str = "",
) -> Tuple[Composed, dict]:
    """Render and add a geneset file to the database.

    This function takes a list of gene values and renders them into a string
    that can be added to the database as a geneset file.

    :param values: The gene values to render.
    :param comments: The comments for the file.
    :return: A query (and params) that can be executed on a cursor.
    """
    contents = "\n".join(f"{value.symbol}\t{value.value}" for value in values)
    size = len(contents)
    return add_geneset_file(size, contents, comments)


def add_geneset_file(
    size: int,
    contents: str,
    comments: str,
) -> Tuple[Composed, dict]:
    """Add a geneset file to the database.

    :param size: The size of the file.
    :param contents: The contents of the file.
    :param comments: The comments for the file.
    :return: A query (and params) that can be executed on a cursor.
    """
    query = (
        SQL("INSERT INTO file")
        + SQL("(file_size, file_contents, file_comments, file_created)")
        + SQL("VALUES (%(file_size)s, %(file_contents)s, %(file_comments)s, NOW())")
        + SQL("RETURNING file_id")
    ).join(" ")
    params = {"file_size": size, "file_contents": contents, "file_comments": comments}
    return query, params



"""Utility functions for the SQL generation functions."""

from typing import Dict, List, Optional, Tuple, Union

from psycopg.sql import SQL, Composed, Identifier, Placeholder

SQLList = List[Union[Composed, SQL]]
ParamDict = Dict[str, Union[str, int]]
OptionalParamDict = Dict[str, Optional[Union[str, int]]]


def construct_filter(
    filters: SQLList,
    params: ParamDict,
    filter_name: str,
    filter_value: Optional[Union[str, int]],
) -> Tuple[SQLList, ParamDict]:
    """Construct a simple filter for a query.

    :param filters: The existing filters.
    :param params: The existing parameters.
    :param filter_name: The filter name to construct.
    :param filter_value: The filter value to construct.

    :return: The constructed filters and parameters.
    """
    if filter_value is not None:
        filters.append(
            SQL("{filter_name} = {param_name}").format(
                filter_name=Identifier(filter_name), param_name=Placeholder(filter_name)
            )
        )
        params[filter_name] = filter_value
    return filters, params


def construct_filters(
    filters: SQLList,
    params: ParamDict,
    filter_items: OptionalParamDict,
) -> Tuple[SQLList, ParamDict]:
    """Construct multiple simple filters for a query.

    Calls the `construct_filter` function for each filter item.

    :param filters: The existing filters.
    :param params: The existing parameters.
    :param filter_items: The filter items to construct.

    :return: The constructed filters and parameters.
    """
    for filter_name, filter_value in filter_items.items():
        filters, params = construct_filter(
            filters,
            params,
            filter_name,
            filter_value,
        )

    return filters, params



"""Functions for querying the user table.

These functions are used to query the user table in the database. They come in two
distinct "flavors": those that return one or more entire user record, and those
that return a single value from a user record. The functions that return one or
more entire user records will return to you the result of calling fetchall() or
fetchone() on the cursor. The functions that return a single value from a user
record will return to you a value extracted from the result of calling fetchone()
or fetchall() on the cursor.

The functions that return one or more entire user records are:
- by_api_key
- by_sso_id
- by_user_id
- by_email

The functions that return a single value from a user record are:
- user_id_from_api_key
- user_id_from_sso_id
"""

from typing import List, Optional

from geneweaver.db.utils import temp_override_row_factory
from psycopg import Cursor, rows


@temp_override_row_factory(rows.tuple_row)
def user_id_from_api_key(cursor: Cursor, api_key: str) -> Optional[int]:
    """Get user id from api key.

    :param cursor: The database cursor.
    :param api_key: The api key to search for.

    :return: The user id (internal) if found, otherwise None.
    """
    cursor.execute(
        """SELECT usr_id FROM production.usr WHERE apikey = %(api_key)s;""",
        {"api_key": api_key},
    )
    result = cursor.fetchone()
    return result[0] if result else None


@temp_override_row_factory(rows.tuple_row)
def user_id_from_sso_id(cursor: Cursor, sso_id: str) -> Optional[int]:
    """Get user id from sso id.

    :param cursor: The database cursor.
    :param sso_id: The sso id to search for.

    :return: The user id (internal) if found, otherwise None.
    """
    cursor.execute(
        """SELECT usr_id FROM production.usr WHERE usr_sso_id = %(sso_id)s;""",
        {"sso_id": sso_id},
    )
    result = cursor.fetchone()
    return result[0] if result else None


def by_api_key(cursor: Cursor, api_key: str) -> List:
    """Get user info by api key.

    :param cursor: The database cursor.
    :param api_key: The api key to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT * FROM production.usr WHERE apikey = %(api_key)s;""",
        {"api_key": api_key},
    )
    return cursor.fetchall()


def by_sso_id(cursor: Cursor, sso_id: str) -> List:
    """Get user info by sso id.

    :param cursor: The database cursor.
    :param sso_id: The sso id to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT * FROM production.usr WHERE usr_sso_id = %(sso_id)s;""",
        {"sso_id": sso_id},
    )
    return cursor.fetchall()


def by_sso_id_and_email(cursor: Cursor, sso_id: str, email: str) -> List:
    """Get user info by sso id and email.

    :param cursor: The database cursor.
    :param sso_id: The sso id to search for.
    :param email: The email to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """
        SELECT * FROM production.usr
        WHERE usr_sso_id = %(sso_id)s AND usr_email = %(email)s;
        """,
        {"sso_id": sso_id, "email": email},
    )
    return cursor.fetchall()


def by_user_id(cursor: Cursor, user_id: int) -> List:
    """Get user info by user id.

    :param cursor: The database cursor.
    :param user_id: The user id (internal) to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT * FROM production.usr WHERE usr_id = %(user_id)s;""",
        {"user_id": user_id},
    )
    return cursor.fetchall()


def by_email(cursor: Cursor, email: str) -> List:
    """Get user info by email.

    :param cursor: The database cursor.
    :param email: The email to search for.

    :return: list of results using `.fetchall()`
    """
    cursor.execute(
        """SELECT * FROM production.usr WHERE usr_email = %(email)s;""",
        {"email": email},
    )
    return cursor.fetchall()


@temp_override_row_factory(rows.tuple_row)
def email_exists(cursor: Cursor, email: str) -> bool:
    """Check if email exists.

    :param cursor: The database cursor.
    :param email: The email to check.

    :return: True if the email exists, otherwise False.
    """
    cursor.execute(
        """ SELECT count(*) FROM usr
            WHERE usr_email = %s
        """,
        (email,),
    )
    existing = int(cursor.fetchone()[0])

    return existing > 0


@temp_override_row_factory(rows.tuple_row)
def sso_id_exists(cursor: Cursor, sso_id: str) -> bool:
    """Check if sso id exists.

    :param cursor: The database cursor.
    :param sso_id: The sso id to check.

    :return: True if the sso id exists, otherwise False.
    """
    cursor.execute(
        """ SELECT count(*) FROM usr
            WHERE usr_sso_id = %s
        """,
        (sso_id,),
    )
    existing = int(cursor.fetchone()[0])

    return existing > 0


@temp_override_row_factory(rows.tuple_row)
def link_user_id_with_sso_id(cursor: Cursor, user_id: int, sso_id: str) -> int:
    """Link a user id with an sso id.

    :param cursor: The database cursor.
    :param user_id: The user id (internal) to link.
    :param sso_id: The sso id to link.

    :return: The user id (internal) of the user that was linked.

    :raises ValueError: If the sso id is already linked to a different account.
    """
    if sso_id_exists(cursor, sso_id):
        raise ValueError("SSO ID is already linked to a different account")

    cursor.execute(
        """UPDATE usr
       SET usr_sso_id = %s
       WHERE usr_id = %s
       RETURNING usr_id;
       """,
        (sso_id, user_id),
    )
    cursor.connection.commit()
    return cursor.fetchone()[0]


@temp_override_row_factory(rows.tuple_row)
def create_sso_user(cursor: Cursor, name: str, email: str, sso_id: str) -> int:
    """Create a new user with sso id.

    :param cursor: The database cursor.
    :param name: The user's name.
    :param email: The user's email address.
    :param sso_id: The user's sso id.

    :return: The user id (internal) of the newly created user.
    """
    split_name = name.split(" ")
    first_name = split_name[0]
    last_name = " ".join(split_name[1:])

    cursor.execute(
        """INSERT INTO usr
           (usr_first_name, usr_last_name,
           usr_email, usr_admin, usr_sso_id,
           usr_last_seen, usr_created, is_guest)
           VALUES
           (%(user_first_name)s, %(user_last_name)s,
           %(user_email)s, '0', %(user_sso_id)s,
           NOW(), NOW(), 'f')
           RETURNING usr_id;
        """,
        {
            "user_first_name": first_name,
            "user_last_name": last_name,
            "user_email": email.lower(),
            "user_sso_id": sso_id,
        },
    )
    cursor.connection.commit()

    return cursor.fetchone()[0]
